#!/usr/bin/env python
# Copyright 2013 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Distutils installer for MAAS."""

# unicode_literals is *not* included here to *better* support setup
# scripts that work on both Python 2 and 3. Both expect their default
# string type, and are fussy about it.
from __future__ import (
    absolute_import,
    print_function,
    )

__version__ = "1.4"
__metaclass__ = type


from codecs import open
from contextlib import contextmanager
from distutils import (
    errors,
    log,
    )
from inspect import getsourcefile
from os import environ

import setuptools


@contextmanager
def envvar(name, value):
    """Context manager for temporarily setting an environment variable."""
    original = environ.get(name)
    environ[name] = value
    try:
        yield
    finally:
        if original is None:
            environ.pop(name, None)
        else:
            environ[name] = original


class install_deps(setuptools.Command):
    """Distutils/setuptools command to install package dependencies."""

    user_options = [
        ('list-from=', None,
         "file from which to read packages to install"),
    ]

    def initialize_options(self):
        self.list_from = "packages.txt"

    def finalize_options(self):
        self.ensure_filename("list_from")
        with open(self.list_from, "r", encoding="utf-8") as fd:
            self.packages = {line.strip() for line in fd}
            self.packages.discard(u"")

    def run(self):
        if len(self.packages) > 0:
            log.info("checking system dependencies")
            self.apt_install_packages()

    def get_missing_packages(self):
        from apt import Cache
        cache = Cache()
        for package in self.packages:
            package = cache[package].mark_install()
        return sorted(self.packages.intersection(
            package.name for package in cache.get_changes()))

    def apt_install_packages(self):
        """Use ``apt`` to check for package installs.

        Spawns ``apt-get`` if necessary.
        """
        missing = self.get_missing_packages()
        if len(missing) == 0:
            log.info("system dependencies are all in place")
        else:
            install = ["apt-get", "-y", "--no-install-recommends", "install"]
            with envvar("DEBIAN_FRONTEND", "noninteractive"):
                self.spawn(["sudo"] + install + missing)


def check_settings(settings):
    """Check setup settings for obvious issues.

    - ``data_files`` should not be specified; setuptools ignores it.

    - ``include_package_data`` should not be specified; it has weird
      behaviour in setuptools.
    """
    if "data_files" in settings:
        raise errors.DistutilsOptionError(
            "data_files is ignored by setuptools")
    if "include_package_data" in settings:
        raise errors.DistutilsOptionError(
            "include_package_data does not work as you "
            "might expect with setuptools")


def setup(**kwargs):
    """MAAS-specific setup function.

    Determines sensible defaults for many parameters.
    """
    command_classes = {}

    try:
        import apt
    except ImportError:
        log.warn("apt not available (install 'python-apt'?)")
    else:
        command_classes["install_deps"] = install_deps
        apt  # Silence lint.

    settings = {
        "cmdclass": command_classes,
        "packages": setuptools.find_packages(".", exclude={"common"}),
        "version": __version__,
        "maintainer": "Canonical",
        "maintainer_email": "maas-devel@lists.launchpad.net",
        "url": "https://launchpad.net/maas",
        "license": "AGPLv3",
        "test_loader": "maastesting.loader:MAASTestLoader",
        "test_suite": ".",
        "tests_require": {"maas-testing >= %s" % __version__},
    }
    settings.update(kwargs)
    check_settings(settings)

    return setuptools.setup(**settings)


def configure(**settings):
    """Create a new ``setup.py`` file in the currnet directory.

    Uses the file in which this function is originally defined as a
    template, adding on a call to ``setup`` with the given ``settings``.
    """
    check_settings(settings)
    thisfile = getsourcefile(configure)
    with open(thisfile, "r", encoding="utf-8") as fin:
        template = fin.read()
    with open("setup.py", "w", encoding="utf-8") as fout:
        fout.write(template)
        fout.write(b"\n\nsetup(**%r)\n" % settings)


setup(**{'install_requires': ['celery >= 2.5.3', 'formencode >= 1.2.4', 'lxml >= 3.2.0', 'maas-apiclient >= 1.4', 'netaddr >= 0.7.7', 'netifaces >= 0.8', 'oops >= 0.0.10', 'oops-datedir-repo >= 0.0.15', 'oops-twisted >= 0.0.6', 'PyYAML >= 3.1.0', 'simplejson >= 3.3.0', 'tempita >= 0.5.1', 'twisted >= 13.0.0', 'txamqp >= 0.6.1', 'zope.interface >= 4.0.5'], 'entry_points': {'console_scripts': set(['maas-cluster-worker = maascluster.entrypoints:celeryd', 'maas-cluster-twistd = maascluster.entrypoints:twistd', 'maas-probe-dhcp = maascluster.entrypoints:probe_dhcp', 'maas-provision = maascluster.entrypoints:provision'])}, 'name': 'maas-cluster', 'tests_require': ['fixtures >= 0.3.8', 'maas-testing >= 1.4', 'mock >= 1.0b1', 'rabbitfixture >= 0.3.4', 'testtools >= 0.9.32'], 'description': 'MAAS Provisioning Server (part of a Cluster Controller).'})
